#!/usr/bin/ruby -w

require 'gtk2'
require 'cairo'
require 'libglade2'
require 'drb'
require 'rinda/ring'
require 'zlib'
require 'rexml/document'

begin
    Dir.chdir('/usr/lib/scotty-tng/client')
rescue
end
require 'sound.rb'

begin
    Dir.chdir('/usr/share/scotty-tng/client')
rescue
end

class Node
  attr_reader :x, :y

  @@nodes = Hash.new

  def initialize(node, layout)
    @name = node.attributes["name"]
    @@nodes[@name] = self

    @sx = @x = node.attributes["x"].to_i
    @sy = @y = node.attributes["y"].to_i

    @vb = Gtk::VBox.new(false, 2)
    if node.attributes['icon'] == nil then
	@icon = nil
    else
	@icon = File.basename(node.attributes['icon'].downcase)
    end
    @img = Gtk::Image.new
    @status = Hash.new
    @nstat = ""

    @vb.pack_start(@img)

    @eb = Gtk::EventBox.new

    @label = Gtk::Label.new
    @label.use_markup = true
    @label.markup = "<small>#{@name}</small>"
    @eb.add(@label)
    @vb.pack_start(@eb)
    layout.put(@vb, @x, @y)
    @vb.signal_connect("size_allocate") { |widget, alloc|
      nx = @x - @vb.size_request[0]/2
      ny = @y - @img.size_request[1]/2

      if @sx != nx or
          @sy != ny then
        @sx = nx
        @sy = ny
        layout.move(widget, nx, ny)
      end
    }
    setStatus
    @vb.show_all
  end

  def updateTooltip
    tt = "<b>#{@name}</b>\n"

    ttmax = 0
    @status.each_pair{|c, s|
	ttmax = Chart.getChart(c).opts['title'].length if ttmax < Chart.getChart(c).opts['title'].length
    }

    @status.sort.each{|c, s|
        tt += "\n<tt>" + sprintf("%#{ttmax}s: ", Chart.getChart(c).opts['title'])
        tt += "<b><span foreground='" + Status.text_color(s) + "'>"
        tt += s.upcase + "</span></b></tt>"
    }

    begin
	# ruby-gnome2-0.16.0-1 for win lacks tooltips
	@vb.tooltip_markup = tt.chomp
    rescue
    end
  end

  def setStatus(sensor=nil, status=nil)
    if sensor != nil and status != nil then
	@status[sensor] = status
    end

    nstat = "ok"
    @status.each_pair{|c, s|
	nstat = s if Status.int(s) > Status.int(nstat)
    }

    if nstat != @nstat then
	@nstat = nstat
	@eb.modify_bg(Gtk::STATE_NORMAL, Gdk::Color.parse(Status.text_color(@nstat)))
	Sound.play("wav/#{@nstat}.wav")
    end

    if @icon != nil then
	fn = @icon + Status.icon_color(nstat)
    else
	fn = "geng"
    end

    if File.readable?("res/#{fn}.png") then
	@img.set("res/#{fn}.png")
    else
	@img.set("res/geng.png")
    end

    updateTooltip
  end

  def self.getNode(name)
    @@nodes[name]
  end
end

class Logger
  def self.setClient(client)
    @@client = client
  end

  def self.logtxt(text, mayiter=true, tag=nil)
    @@client.logtxt(text, mayiter, tag)
  end
end

class Chart
  attr_reader :x, :y, :opts

  CWidth = 60
  CHeight = 35
  CFoot = 10

  @@charts = Hash.new
  @@tooltips = Hash.new

  def initialize(chart, layout)
    @nname = chart.attributes["node"]
    @sname = chart.attributes["sensor"]
    @lname = @nname + "-" + @sname
    @@charts[@lname] = self

    Logger.logtxt("Retrieving chart '#{@lname}'")
    begin
      @opts = Client.getTS.read([:DRingLOpts, @lname, nil], 0)[2]
    rescue
      Logger.logtxt("Could not get chart: #{$!}")
      return
    end
    @img = Cairo::ImageSurface.new(Cairo::FORMAT_ARGB32, CWidth, CHeight + CFoot)
    @ct = Cairo::Context.new(@img)

    @da = Gtk::DrawingArea.new
    @da.set_size_request(CWidth, CHeight + CFoot)
    @da.signal_connect("expose_event") { |widget, arg0|
      draw_chart_data(@ct)

      # Copy image to drawing area surface
      c = widget.window.create_cairo_context
      c.set_source(@img, 0, 0)
      c.paint
    }

    @x = chart.attributes["x"].to_i
    @y = chart.attributes["y"].to_i
    @sx = chart.attributes["x"].to_i
    @sy = chart.attributes["y"].to_i
    @data = Array.new(@opts['lines']),
    @current = Array.new(@opts['lines'], 0)
    @show_current = Array.new(@opts['lines'], true)
    @last = Array.new(@opts['lines'], 0)
    @sum = Array.new(@opts['lines'], 0)
    @max = Array.new(@opts['lines'], 0)
    @status = [nil, nil, nil]

    0.upto(@opts['lines'] - 1){ |l|
      @data[l] = Array.new
    }

    layout.put(@da, @x, @y)
    @da.signal_connect("size_allocate") { |widget, alloc|
      nx = @x - alloc.width/2
      ny = @y - alloc.height/2

      if @sx != nx or
          @sy != ny then
        @sx = nx
        @sy = ny
        layout.move(widget, nx, ny)
      end
    }
    @da.show_all

    # Read current status
    stat = [Time.now.to_i, "ok", nil]
    status = Client.getTS.read_all(["SensorStatus", @lname, nil])
    status.each{ |s|
      stat = s[2]
    }
    setStatus(stat, true)
  end

  def setStatus(status, startup=false)
    if status[1] != @status[1] or status[2] != @status[2] then
	if status[1] != "ok" or not startup then
	    msg = "[#{Time.at(status[0]).strftime('%X')}] #{@opts['title']} "
	    msg += "[#{@opts['descr']}] " if @opts['descr']
	    msg += "@ #{@nname} changed to #{status[1].upcase}"
	    msg += ": #{status[2]}" if status[2]
	    Logger.logtxt(msg, false, status[1])
	    Chart.updSIcon(self, status)
	end

	@status = status
	Node.getNode(@nname).setStatus(@lname, status[1]) if Node.getNode(@nname)
	updTooltip
    end
  end

  def self.updSIcon(chart, status)
    if status[1] == "ok" then
	@@tooltips.delete(chart)
    else
	@@tooltips[chart] = status
    end

    tooltip = "Scotty TNG\n"

    stat = "ok"
    @@tooltips.each_pair{|k, s|
	tooltip += "\n#{k.opts['title']}: #{s[2]}"

	stat = s[1] if Status.int(s[1]) > Status.int(stat)
    }

    case stat
    when "ok"
	Client.getSIcon.stock = Gtk::Stock::DIALOG_INFO
    when "info"
	Client.getSIcon.stock = Gtk::Stock::INFO
    when "warning"
	Client.getSIcon.stock = Gtk::Stock::DIALOG_WARNING
    when "critical"
	Client.getSIcon.stock = Gtk::Stock::DIALOG_ERROR
    end
    Client.getSIcon.tooltip = tooltip
    Client.getSIcon.blinking = (stat == "critical")
  end

  def pushVal(dval)
    0.upto(@opts["lines"] - 1) do |l|
      @show_current[l] = (dval[l] != nil)
      dval[l] = 0 unless @show_current[l]

      case @opts["type"][l]
      when "c"
        if @current[l] - dval[l] > 0 then
          val = (@current[l] - dval[l]) / (Time.now.to_f - @last)
        else
          val = 0
        end
      else
        val = dval[l]
      end
      
      @current[l] = dval[l]
      @data[l].push(val)
      @sum[l] += val
      
      if @data[l].size > CWidth then
        @sum[l] -= @data[l].shift
      end
      
      @max[l] = val*1.2 if @max[l] == nil or @max[l] < val
    end
      
    @last = Time.now.to_f
    @da.queue_draw
    updTooltip
  end

  def updTooltip
    tt  = "<b>#{@opts['title']} on #{@nname}</b>\n"
    tt += "<small>#{@opts['descr']}</small>\n" if @opts['descr'] != nil

    ttmax = 6
    0.upto(@opts["lines"] - 1) do |l|
      ttmax = @opts['legend'][l].length if ttmax < @opts['legend'][l].length
    end

    tt += "<tt>" + sprintf("\n%#{ttmax}s: ", "Status") 
    tt += "<b><span foreground='" + Status.text_color(@status[1]) + "'>"
    tt += @status[1].upcase + "</span></b></tt>\n"
    tt += "<small>" + @status[2] + "</small>\n" if @status[2]
    
    0.upto(@opts["lines"] - 1) do |l|  
      next if @opts["legend"][l] == nil

      tt += "\n<tt><span foreground='" + @opts['color'][l] + "'>"
      tt += sprintf("%#{ttmax}s", @opts['legend'][l]) + ": "
      tt += Client.format_data(@opts['format'][l], @current[l])
      tt += "</span></tt>"
    end

    begin
      # ruby-gnome2-0.16.0-1 for win lacks tooltips
      @da.tooltip_markup = tt.chomp
    rescue
    end
  end

  # Draws the content of a chart
  def draw_frame(ct)
    Client.set_ct_scol(ct, Status.text_color(@status[1]))

    ct.rectangle(0, 0, CWidth, CHeight)
    ct.move_to(1, 9)
    ct.show_text(@opts["title"])
    ct.stroke

    ct.set_source_rgb(0.9, 0.9, 0.9)
    ct.rectangle(0, CHeight, CWidth, CFoot)
    ct.fill_preserve
    ct.stroke
  end

  def draw_chart_data(ct)
    height = CHeight
    width = CWidth

    ct.line_width = 1
    ct.set_font_size(9);

    # Clear chart
    ct.set_source_rgb(1, 1, 1)
    ct.rectangle(0, 0, width, height)
    ct.fill_preserve
    ct.stroke

    # Get maximum values (dynamic for scaling)
    totalmax = 0
    0.upto(@opts['lines'] - 1) do |l|
	totalmax = @max[l] if @max[l] > totalmax
    end

    # Draw the lines
    hasdata = 0
    0.upto(@opts['lines'] - 1) do |l|
	next if @data[l].length == 0
	hasdata = 1
	next if @opts['slegend'][l] == nil

	Client.set_ct_scol(ct, @opts['color'][l])

	if @opts['max'][l] > 0 then
	    max = totalmax
	else
	    max = @max[l]
	end

	max = 1 if max == 0

	first = 1
	x = 1
	@data[l].each{|p|
	    begin
		y = height - p * (height - 10) / max
		y = 1 if y < 1
	    rescue
		y = height - 1
	    end

	    if first == 1 then
		if l == 0 then
		    ct.move_to(x, height)
		    ct.line_to(x, y)
		else
		    ct.move_to(x, y)
		end
		first = 0
	    else
		ct.line_to(x, y)
	    end
	    x += 1
	}
	if l == 0 then
	    ct.line_to(x-1, height)
	    ct.fill
	end
	ct.stroke
    end

    # Warn about missing data
    if hasdata == 0 then
	draw_frame(ct)

	ct.set_font_size(8);
	ct.set_source_rgb(0.8, 0, 0)

	ct.move_to(1, Chart::CHeight + Chart::CFoot - 2)
	ct.show_text("No data!")

	return
    end
    
    ct.set_dash([2], 0)
    0.upto(@opts['lines'] - 1) do |l|
	next if @opts['slegend'][l] == nil

	begin
	    Client.set_ct_scol(ct, "dark" + @opts['color'][l])

	    if @opts['max'][l] > 0 then
		max = totalmax
	    else
		max = @max[l]
	    end

	    max = 1 if max == 0

	    avg = @sum[l]/@data[l].size
	    y = height - avg * (height - 10) / max

	    ct.move_to(      0, y)
	    ct.line_to(width-1, y)
	    ct.stroke
	rescue
	    p $!
	end
    end
    ct.set_dash(nil, 0)

    # Draw frame
    draw_frame(ct)

    # Print current values
    ct.set_font_size(8);
    ct.move_to(1, CHeight + CFoot - 2)

    0.upto(@opts['lines'] - 1) do |l|
	if @opts['slegend'][l] != nil and
	   @show_current[l] then
	    Client.set_ct_scol(ct, "dark" + @opts['color'][l])
	    ct.show_text(Client.format_data(@opts['format'][l], @data[l].last) + " ")
	end
    end
    ct.stroke
  end

  def self.getChart(lname)
    @@charts[lname]
  end
end

class Status
  def self.text_color(status)
    case status
    when "critical"
      return "red"
    when "warning"
      return "orange"
    when "info"
      return "goldenrod"
    when "ok"
      return "green"
    else
      return "goldenrod"
    end
  end

  def self.icon_color(status)
    case status
    when "critical"
      return "rd"
    when "warning"
      return "og"
    when "info"
      return "yw"
    when "ok"
      return "gn"
    else
      return "yw"
    end
  end

  def self.int(status)
    case status
    when "critical"
	return 4
    when "warning"
	return 3
    when "info"
	return 2
    when "ok"
	return 1
    else
	return 0
    end
  end

  def self.comp(ns, os)
    n = self.int(ns)
    o = self.int(os)

    return n if n > o
    return o
  end
end

class Client
  include GetText

  attr :glade

  def self.getTS
    @@ts
  end

  def self.getSIcon
    @@sicon
  end

  def initialize(path_or_data, root = nil, domain = nil, localedir = nil, flag = GladeXML::FILE)
    bindtextdomain(domain, localedir, nil, "UTF-8")
    Logger.setClient(self)
    @glade = GladeXML.new(path_or_data, root, domain, localedir, flag) {|handler| method(handler)}
    @@sicon = Gtk::StatusIcon.new
    @@sicon.stock = Gtk::Stock::DIALOG_INFO
    @window_loading = @glade.get_widget("window_loading")
    @window_main = @glade.get_widget("window_main")
    @scrolledwindow_log = @glade.get_widget("scrolledwindow_log")

    @textview_log = @glade.get_widget("textview_log")
    ["ok", "info", "warning", "critical"].each{|l|
	tag = Gtk::TextTag.new(l)
	tag.foreground = Status.text_color(l)
	@textview_log.buffer.tag_table.add(tag)
    }

    @dialog_selview = @glade.get_widget("dialog_selview")
    @dialog_selring = @glade.get_widget("dialog_selring")
    @menu = @glade.get_widget("menu")

    # Show main window and startup splash
    Gtk::Window.auto_startup_notification = false
    image_loading = @glade.get_widget("image_loading")
    image_loading.set_file("res/scotty-tng.png")
    @window_main.show
    @window_loading.show
    Gtk::Window.auto_startup_notification = true

    @width = 320	# initial width of the view
    @height = 240	# initial height of the view

    # Gtk::Layout is used to draw the view
    @layout = @glade.get_widget("layout_main")
    @layout_img = nil
    @layout_ct = nil
  end

  # Process pending Gtk events and return
  def iter
    while Gtk.events_pending? do
	Gtk.main_iteration
    end
  end

  # Print a loging message on screen
  def logtxt(text, mayiter=true, tag=nil)
    if tag
	@textview_log.buffer.insert(@textview_log.buffer.end_iter, text.to_s + "\n", tag)
    else
	@textview_log.buffer.insert(@textview_log.buffer.end_iter, text.to_s + "\n")
    end

    iter if mayiter

    @scrolledwindow_log.vadjustment.value = @scrolledwindow_log.vadjustment.upper -
	@scrolledwindow_log.vadjustment.page_size
    iter if mayiter
  end

  # Initialize the DRb stuff
  def init_drb
    iter

    logtxt("Starting DRb service")

    DRb.start_service
    rf = Rinda::RingFinger.new
    doloop = 1

    while doloop == 1
      logtxt("Lookup ring servers")
      @rs = Hash.new
      rf.lookup_ring { |s|
	@rs[s.__drburi] = s
      }
      
      if @rs.size == 0 then
	logtxt("Error: Could not find any ring server!")
	dialog = Gtk::Dialog.new("Error",
				 @window_loading,
				 Gtk::Dialog::MODAL,
				 [Gtk::Stock::REFRESH, Gtk::Dialog::RESPONSE_ACCEPT],
				 [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_REJECT]
				 )
	dialog.vbox.add(Gtk::Label.new("Could not find ring server!"))
        dialog.show_all

        if dialog.run == Gtk::Dialog::RESPONSE_ACCEPT then
	  dialog.destroy
	  next 
	end

        return 0
      elsif @rs.size > 1 then
	  @dialog_selring.set_modal(true)
	  @dialog_selring.show_all
	  exit if @dialog_selring.run != Gtk::Dialog::RESPONSE_ACCEPT

	  doloop = 0
      else
	  @@ts = @rs.values.shift
	  logtxt("Using #{@@ts.__drburi}")

	  doloop = 0
      end
    end


    @dialog_selview.show_all
    exit if @dialog_selview.run != Gtk::Dialog::RESPONSE_ACCEPT

    @window_loading.hide

    return 1
  end

  # Load a view from XML string
  def load_view(xview)
    @window_main.title = xview.attributes["title"] + " [Scotty TNG]"

    # Setup layout size
    @width = xview.attributes["width"].to_i
    @height = xview.attributes["height"].to_i

    @layout.set_size(@width, @height)
    @layout_img = Cairo::ImageSurface.new(Cairo::FORMAT_ARGB32, @width, @height)
    @layout_ct = Cairo::Context.new(@layout_img)

    # Add text elements
    xview.elements.each("text") { |text|
	label = Gtk::Label.new
	if text.attributes["color"] != nil then
	    label.markup = "<span foreground='"+ text.attributes['color'] +"'>#{text.text}</span>"
	else
	    label.text = text.text
	end
	@layout.put(label, text.attributes["x"].to_i, text.attributes["y"].to_i)
	label.show
    }

    # Add node elements
    xview.elements.each("node") { |node|
        Node.new(node, @layout)
    }

    # Add chart elements
    xview.elements.each("chart") { |chart|
      Chart.new(chart, @layout)
    }

    
    # Record graph updates
    @obsthread = Thread.start do
      observer = @@ts.notify 'write', [:DRingLVals, nil, nil]
      observer.each{|e, t|
        dname = t[1]
        dval = t[2]

        chart = Chart.getChart(dname)

        chart.pushVal(dval) if chart
      }
    end

    # Record status changes
    @ostatsthread = Thread.start do
	obsstatus = @@ts.notify 'write', ["SensorStatus", nil, nil]
	obsstatus.each{|e, t|
	    dname = t[1]
            status = t[2]
	    chart = Chart.getChart(dname)
	    chart.setStatus(status) if chart
	}
    end

    # Draw links between nodes on background
    xview.elements.each("link") { |link|
	Client.set_ct_scol(@layout_ct, link.attributes["color"].to_s)
	n1 = Node.getNode(link.attributes["n1"])
	n2 = Node.getNode(link.attributes["n2"])

	if n1 != nil and n2 != nil then
	    @layout_ct.move_to(n1.x, n1.y)
	    @layout_ct.line_to(n2.x, n2.y)
	    @layout_ct.stroke
	end
    }
    @layout.queue_draw
  end


  ##
  ## window_main
  ##

  # Set cairo context color by string
  def self.set_ct_scol(ct, color)
    begin
        ct.set_source_color(color)
    rescue
	case color.downcase
	    when "red"
		ct.set_source_rgb(1, 0, 0)
	    when "darkred"
		ct.set_source_rgb(0.7, 0, 0)
	    when "green"
		ct.set_source_rgb(0, 1, 0)
	    when "darkgreen"
		ct.set_source_rgb(0, 0.7, 0)
	    when "blue"
		ct.set_source_rgb(0, 0, 1)
	    when "darkblue"
		ct.set_source_rgb(0, 0, 0.7)
	    else
		# Nothing matched, black is always good!
		ct.set_source_rgb(0, 0, 0)
	end
    end
  end

  # Formats a value for a given format (=[t]ime, [p]ercent, [b]ytes...)
  def self.format_data(format, val)
    case format
	# Time in ms
	when "t"
	    return sprintf("%dms", val) if val < 1000
	    return sprintf("%.1fs", val/1000)
	# Percent
	when "p"
	    return sprintf("%d%%", val)
	# Bitrate
	when "r"
	    val *= 8
	    
	    return sprintf("%.1fT", val / 1000000000000.0) if val > 100000000000
	    return sprintf("%.1fG", val / 1000000000.0   ) if val > 100000000
	    return sprintf("%.1fM", val / 1000000.0      ) if val > 100000
	    return sprintf("%.1fk", val / 1000.0         ) if val > 100
	    return sprintf("%d", val)
	when "b"
	    return "yes" if val > 0
	    return "no"
	# Anything else
	else
	    return sprintf("%d", val)
    end
  end

  # Draw layout background (link lines)
  def layout_main_expose_event_cb(widget, arg0)
    if @layout_img then
	c = widget.bin_window.create_cairo_context
	c.set_source(@layout_img, 0, 0)
	c.paint
    end

    return false
  end

  # Terminate if main window is closed
  def window_main_delete_event_cb(widget, arg0)
    Gtk.main_quit
  end

  def gtk_main_quit(widget)
    Gtk.main_quit
  end

  # Popup menu
  def layout_main_button_press_event_cb(widget, arg0)
    @menu.popup(nil, nil, arg0.button, arg0.time) if arg0.button == 3
  end

  # Open
  def menuitem_open_activate_cb(widget)
    puts "menuitem_open_activate_cb() is not implemented yet."
  end

  ##
  ## dialog_selring
  ##

  def dialog_selring_show_cb(widget)
    tview = @glade.get_widget("treeview_ring")

    liststore = Gtk::ListStore.new(String)

    renderer = Gtk::CellRendererText.new
    col = Gtk::TreeViewColumn.new("RingServer", renderer)
    col.set_cell_data_func(renderer) do |col, renderer, model, iter|
	renderer.text = iter[0]
    end

    tview.selection.mode = Gtk::SELECTION_SINGLE
    tview.model = liststore
    tview.append_column(col)

    @rs.values.each{|views|
      viewset = liststore.append
      viewset[0] = views.__drburi
    }
  end


  ##
  ## dialog_selview
  ##
  def dialog_selview_show_cb(widget)
    viewsets = @@ts.read_all([:ScottyTNGd, nil, nil])
    tview = @glade.get_widget("treeview_views")

    treestore = Gtk::TreeStore.new(String, REXML::Document)

    renderer = Gtk::CellRendererText.new
    col = Gtk::TreeViewColumn.new("View", renderer)
    col.set_cell_data_func(renderer) do |col, renderer, model, iter|
	renderer.markup = iter[0]
    end

    tview.selection.mode = Gtk::SELECTION_SINGLE
    tview.model = treestore
    tview.append_column(col)

    viewsets.each{|views|
	viewset = treestore.append(nil)
	viewset[0] = "<b>" + views[1] + "</b>"
	viewset[1] = nil
	
	xml = REXML::Document.new( Zlib::Inflate.inflate(views[2]) )
	xml.elements.each("/views/view") { |xview|
	    view = treestore.append(viewset)
	    view[0] = xview.attributes["title"]
	    view[1] = xview
	}
    }

    tview.expand_all
  end

  def treeview_views_row_activated_cb(widget, arg0, arg1)
    dialog = widget.toplevel
    case dialog.name
      when "dialog_selview"
      tview = @glade.get_widget("treeview_views")

      it = tview.model.get_iter(arg0)
      if it.get_value(1) != nil then
	dialog.hide
        iter
	load_view(it.get_value(1))
	dialog.response(Gtk::Dialog::RESPONSE_ACCEPT)
      end
      when "dialog_selring"
      tview = @glade.get_widget("treeview_ring")

      it = tview.model.get_iter(arg0)
      if it.get_value(0) != nil then
	@ts = @rs[it.get_value(0)]
	dialog.hide
	dialog.response(Gtk::Dialog::RESPONSE_ACCEPT)
      end
      logtxt("Using #{@@ts.__drburi}")
      else
      return
    end
  end

  def button_exit_clicked_cb(widget)
    dialog = widget.toplevel
    dialog.hide
  end
end

Gtk.main if Client.new("client.glade", nil, "Scotty TNG").init_drb
